#include "tetris.h"
#include <shellapi.h>
#include <windows.h>

/*
 * global variables
 */

// link with shell32.lib


int score; // current score

char* keys[] = { // main menu accelerator names
	"Enter", "Esc", "Space", "Pg Up", "Pg Down", "End", "Home",
	"Left", "Up", "Right", "Down", "Insert", "Delete",
	"0", "1", "2", "3", "4", "5", "6", "7", "8", "9",
	"A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M",
	"N", "O", "P", "Q", "R", "S", "T", "U", "V", "W", "X", "Y", "Z",
	"Num 0", "Num 1", "Num 2", "Num 3", "Num 4", "Num 5", "Num 6", "Num 7",
	"Num 8", "Num 9", "F1", "F2", "F3", "F4", "F5", "F6", "F7", "F8", "F9",
	"F10", "F11", "F12"
};
ACCEL defaultAccels[NUM_ACCELS] = { // factory default accelerators
	{ FVIRTKEY | FNOINVERT, VK_F2, IDM_GAME_NEW },
	{ FVIRTKEY | FNOINVERT, VK_SPACE, IDM_GAME_PAUSE },
	{ FVIRTKEY | FNOINVERT | FCONTROL, VK_H, IDM_GAME_HIGHSCORES },
	{ FVIRTKEY | FNOINVERT, VK_ESCAPE, IDM_GAME_EXIT },
	{ FVIRTKEY | FNOINVERT, VK_LEFT, IDM_CONTROL_LEFT },
	{ FVIRTKEY | FNOINVERT, VK_RIGHT, IDM_CONTROL_RIGHT },
	{ FVIRTKEY | FNOINVERT, VK_DOWN, IDM_CONTROL_DOWN },
	{ FVIRTKEY | FNOINVERT, VK_UP, IDM_CONTROL_ROTATE },
	{ FVIRTKEY | FNOINVERT | FCONTROL, VK_LEFT, IDM_CONTROL_FORCE_LEFT },
	{ FVIRTKEY | FNOINVERT | FCONTROL, VK_RIGHT, IDM_CONTROL_FORCE_RIGHT },
	{ FVIRTKEY | FNOINVERT, VK_NEXT, IDM_CONTROL_FORCE_DOWN },
	{ FVIRTKEY | FNOINVERT, VK_1, IDM_OPTIONS_SPEED_1 },
	{ FVIRTKEY | FNOINVERT, VK_2, IDM_OPTIONS_SPEED_2 },
	{ FVIRTKEY | FNOINVERT, VK_3, IDM_OPTIONS_SPEED_3 },
	{ FVIRTKEY | FNOINVERT, VK_4, IDM_OPTIONS_SPEED_4 },
	{ FVIRTKEY | FNOINVERT | FCONTROL | FALT, VK_B, IDM_OPTIONS_COLOR_BACK },
	{ FVIRTKEY | FNOINVERT | FCONTROL | FALT, VK_L, IDM_OPTIONS_COLOR_BLOCK },
	{ FVIRTKEY | FNOINVERT | FCONTROL, VK_T, IDM_OPTIONS_WINDOW_TOPMOST },
	{ FVIRTKEY | FNOINVERT | FCONTROL, VK_E, IDM_OPTIONS_WINDOW_EASYMOVE },
	{ FVIRTKEY | FNOINVERT | FCONTROL, VK_K, IDM_OPTIONS_KEYBINDINGS },
	{ FVIRTKEY | FNOINVERT | FCONTROL, VK_N, IDM_OPTIONS_NEXT },
	{ FVIRTKEY | FNOINVERT | FCONTROL, VK_G, IDM_OPTIONS_GRID },
	{ FVIRTKEY | FNOINVERT | FCONTROL, VK_L, IDM_OPTIONS_LAND },
	{ FVIRTKEY | FNOINVERT, VK_F1, IDM_HELP_ABOUT }
};
HACCEL hAccel; // handle to current accelerator table

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdLine,
		int iCmdShow) {
	static char szAppClassName[] = "Tetris";

	if (CreateMutex(NULL, FALSE, szAppClassName)
			&& (GetLastError() == ERROR_ALREADY_EXISTS)) {
		SetForegroundWindow(FindWindow(szAppClassName, NULL));
		return 0;
	}

	WNDCLASS wc;
	wc.style = CS_BYTEALIGNCLIENT | CS_BYTEALIGNWINDOW;
	wc.lpfnWndProc = WndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hIcon = LoadIcon(hInstance, szAppClassName);
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = (HBRUSH) GetStockObject(BLACK_BRUSH);
	wc.lpszMenuName = szAppClassName;
	wc.lpszClassName = szAppClassName;
	RegisterClass(&wc);

	HWND hWnd = CreateWindow(
					szAppClassName,
					szAppClassName,
					WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX,
					0,
					0,
					0,
					0,
					NULL,
					NULL,
					hInstance,
					NULL);
	ShowWindow(hWnd, iCmdShow);

	MSG msg;
	while (GetMessage(&msg, NULL, 0, 0)) {
		if (!TranslateAccelerator(hWnd, hAccel, &msg)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
	return msg.wParam;
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	static int cup[NUM_ROWS][NUM_COLS]; 			// game board
	static int figures[NUM_FIGURES][SIDE][SIDE] = { // available figures
		{ { 0, 1, 0 }, { 0, 1, 0 }, { 0, 1, 0 } },	// |
		{ { 1, 1, 1 }, { 1, 1, 1 }, { 1, 1, 1 } },	// 3x3 solid block
		{ { 0, 0, 1 }, { 1, 1, 1 }, { 0, 0, 0 } },	// L
		{ { 1, 0, 0 }, { 1, 1, 1 }, { 0, 0, 0 } },	// _|
		{ { 1, 1, 1 }, { 0, 1, 0 }, { 0, 1, 0 } },	// T
		{ { 1, 1, 1 }, { 0, 1, 0 }, { 0, 0, 0 } },	// little T
		{ { 1, 1, 1 }, { 1, 0, 1 }, { 0, 0, 0 } },	// n
		{ { 0, 0, 1 }, { 0, 1, 1 }, { 1, 1, 0 } },	// smth like 'w'
		{ { 0, 0, 0 }, { 0, 1, 0 }, { 0, 0, 0 } },	// .
		{ { 1, 1, 0 }, { 0, 1, 1 }, { 0, 0, 0 } }	// s
	};
	static int (*curFigure) [SIDE];
	static int (*nextFigure) [SIDE];
	static int xPos, yPos; 			// current figure position, top-left corner
	static BOOL fPause; 			// pause flag
	static BOOL fSave;	 			// save on exit flag
	static HMENU hMenu;				// handle to main menu
	static RECORDS records[3];
	static int options[] = {
		2,							// game speed (1 - 4)
		RGB(0, 0, 0),				// background color
		RGB(255, 255, 255), 		// block color
		TRUE,						// show next figure flag
		TRUE,						// show grid flag
		TRUE,						// show landspot flag
		FALSE,						// window always on top flag
		TRUE						// window easy move flag
	};
	static ACCEL accels[NUM_ACCELS];	// current accelerator table

	switch (uMsg) {
		case WM_CREATE:
		{
			hMenu = GetMenu(hWnd);
			HANDLE hFile = CreateFile("tetris.dat", GENERIC_READ, FILE_SHARE_READ,
					NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL
					| FILE_FLAG_SEQUENTIAL_SCAN, NULL);
			if (hFile != INVALID_HANDLE_VALUE) {
				DWORD dwBytesRead;
				ReadFile(hFile, options, sizeof options, &dwBytesRead, NULL);
				ReadFile(hFile, records, sizeof records, &dwBytesRead, NULL);
				ReadFile(hFile, accels, sizeof accels, &dwBytesRead, NULL);
				CloseHandle(hFile);
				CheckMenuItem(hMenu, IDM_OPTIONS_SAVE, MF_CHECKED);
				fSave = TRUE;
			}
			else
				memcpy(accels, defaultAccels, NUM_ACCELS * sizeof *defaultAccels);

			hAccel = CreateAcceleratorTable(accels, NUM_ACCELS);

			// center window on screen
			int cxScreen = GetSystemMetrics(SM_CXSCREEN);
			int cyScreen = GetSystemMetrics(SM_CYSCREEN);
			int width = (NUM_COLS * 10) + (GetSystemMetrics(SM_CXFIXEDFRAME) * 2);
			int height = (NUM_ROWS * 10) + (GetSystemMetrics(SM_CYFIXEDFRAME) * 2)
					+ GetSystemMetrics(SM_CYCAPTION) + GetSystemMetrics(SM_CYMENU);

			SetWindowPos(hWnd, options[opTopmost] ? HWND_TOPMOST : HWND_NOTOPMOST,
					(cxScreen - width) / 2, (cyScreen - height) / 2, width, height,
					SWP_NOREDRAW);

			// final touches...
			refreshMenu(hMenu, accels, (BOOL) curFigure, fPause, options);
			SetClassLong(hWnd, GCL_HBRBACKGROUND, (LONG) CreateSolidBrush(options[opBack]));
			return 0;
		}
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDM_GAME_NEW:
					if (score > records[2].score) {
						// it's a record
						fPause = TRUE;
						DialogBoxParam((HINSTANCE) GetWindowLong(hWnd, GWL_HINSTANCE),
								 "HighScores", hWnd, HighScoresProc, (LPARAM) records);
					}
					// reset everything, prepare for new game
					writeScore(hWnd, score = 0, options[opSpeed]);
					memset(cup, 0, sizeof cup);
					srand(GetTickCount());
					curFigure = figures[rand() % NUM_FIGURES];
					// two GetTickCount()s in a row don't differ on a PIII 650MHz ;)
					nextFigure = figures[(GetTickCount() << 4) % NUM_FIGURES];
					putBlockEx(cup, curFigure, xPos = NUM_COLS / 2 - 1, yPos = 0, btActive);
					SetTimer(hWnd, ID_TIMER, 1300 - options[opSpeed] * 300, NULL);
					refreshMenu(hMenu, accels, (BOOL) curFigure, fPause = FALSE, options);
					InvalidateRect(hWnd, NULL, TRUE);
					return 0;
				case IDM_GAME_PAUSE:
					fPause ^= TRUE;
					refreshMenu(hMenu, accels, (BOOL) curFigure, fPause, options);
					if (fPause)
						SetWindowText(hWnd, "--==Paused==--");
					else
						writeScore(hWnd, score, options[opSpeed]);
					return 0;
				case IDM_GAME_HIGHSCORES:
				{
					// read-only access
					int i = score;
					score = 0;
					DialogBoxParam((HINSTANCE) GetWindowLong(hWnd, GWL_HINSTANCE),
							"HighScores", hWnd, HighScoresProc, (LPARAM) records);
					score = i;
					return 0;
				}
				case IDM_GAME_EXIT:
					SendMessage(hWnd, WM_CLOSE, 0, 0);
					return 0;
				case IDM_CONTROL_LEFT:
				{
					for (int x = xPos; x < xPos + SIDE; x++)
						for (int y = yPos; y < yPos + SIDE; y++)
							if (cup[y][x] == 2 && (x == 0 || cup[y][x - 1] == 1))
								// can't move left
								return 0;

					putBlockEx(cup, curFigure, xPos, yPos, btZero);
					putBlockEx(cup, curFigure, --xPos, yPos, btActive);

					RECT rect;
					SetRect(&rect, xPos * 10, yPos * 10,
							(xPos + 4) * 10, (yPos + SIDE) * 10);
					InvalidateRect(hWnd, &rect, TRUE);

					if (options[opLand]) {
						// show landspot
						SetRect(&rect, 0, NUM_ROWS * 10 - 3, NUM_COLS * 10, NUM_ROWS * 10);
						InvalidateRect(hWnd, &rect, TRUE);
					}
					return 0;
				}
				case IDM_CONTROL_RIGHT:
				{
					for (int x = min(xPos + 2, NUM_COLS - 1); x >= xPos; x--)
						for (int y = yPos; y < yPos + SIDE; y++)
							if (cup[y][x] == 2 && (x == NUM_COLS - 1 || cup[y][x + 1] == 1))
								// can't move right
								return 0;

					putBlockEx(cup, curFigure, xPos, yPos, btZero);
					putBlockEx(cup, curFigure, ++xPos, yPos, btActive);

					RECT rect;
					SetRect(&rect, (xPos - 1) * 10, yPos * 10,
							(xPos + SIDE) * 10, (yPos + SIDE) * 10);
					InvalidateRect(hWnd, &rect, TRUE);

					if (options[opLand]) {
						// show landspot
						SetRect(&rect, 0, NUM_ROWS * 10 - 3, NUM_COLS * 10, NUM_ROWS * 10);
						InvalidateRect(hWnd, &rect, TRUE);
					}
					return 0;
				}
				case IDM_CONTROL_DOWN:
					SendMessage(hWnd, WM_TIMER, (WPARAM) ID_TIMER, 0);
					return 0;
				case IDM_CONTROL_ROTATE:
				{
					int x;
					int y;
					for (y = yPos; y < yPos + SIDE; y++) {
						for (x = xPos; x < xPos + SIDE; x++)
							if (cup[y][x] == 1)
								break;
						if (x < xPos + SIDE)
							break;
					}
					if ((xPos < 0) || (xPos >= NUM_COLS - 2)
							|| (yPos >= NUM_ROWS - 2) || (x < xPos + SIDE))
						// can't rotate
						return 0;

					putBlockEx(cup, curFigure, xPos, yPos, btZero);
					x = curFigure[0][0];
					y = curFigure[0][1];
					curFigure[0][0] = curFigure[2][0];
					curFigure[0][1] = curFigure[1][0];
					curFigure[2][0] = curFigure[2][2];
					curFigure[1][0] = curFigure[2][1];
					curFigure[2][2] = curFigure[0][2];
					curFigure[2][1] = curFigure[1][2];
					curFigure[0][2] = x;
					curFigure[1][2] = y;
					putBlockEx(cup, curFigure, xPos, yPos, btActive);

					RECT rect;
					SetRect(&rect, xPos * 10, yPos * 10,
							(xPos + SIDE) * 10, (yPos + SIDE) * 10);
					InvalidateRect(hWnd, &rect, TRUE);

					if (options[opLand]) {
						// show landspot
						SetRect(&rect, 0, NUM_ROWS * 10 - 3, NUM_COLS * 10, NUM_ROWS * 10);
						InvalidateRect(hWnd, &rect, TRUE);
					}
					return 0;
				}
				case IDM_CONTROL_FORCE_LEFT:
					for (int x = xPos; x == xPos; x--)
						SendMessage(hWnd, WM_COMMAND, IDM_CONTROL_LEFT, 0);
					return 0;
				case IDM_CONTROL_FORCE_RIGHT:
					for (int x = xPos; x == xPos; x++)
						SendMessage(hWnd, WM_COMMAND, IDM_CONTROL_RIGHT, 0);
					return 0;
				case IDM_CONTROL_FORCE_DOWN:
					if (!fPause)
						while (yPos)
							SendMessage(hWnd, WM_TIMER, (WPARAM) ID_TIMER, 0);
					return 0;
				case IDM_OPTIONS_SPEED_1:
				case IDM_OPTIONS_SPEED_2:
				case IDM_OPTIONS_SPEED_3:
				case IDM_OPTIONS_SPEED_4:
					CheckMenuRadioItem(hMenu, IDM_OPTIONS_SPEED_1,
							IDM_OPTIONS_SPEED_4, wParam, MF_BYCOMMAND);
					options[opSpeed] = LOWORD(wParam) - IDM_OPTIONS_SPEED_1 + 1;
					if (curFigure) {
						// set timer only if game started
						SetTimer(hWnd, ID_TIMER, 1300 - options[opSpeed] * 300, NULL);
						if (!fPause)
							writeScore(hWnd, score, options[opSpeed]);
					}
					return 0;
				case IDM_OPTIONS_COLOR_BACK:
				{
					static COLORREF custColors[] = {
						RGB(0, 0, 0), RGB(0, 0, 0x80), RGB(0, 0x80, 0),
						RGB(0, 0x80, 0x80), RGB(0x80, 0, 0), RGB(0x80, 0, 0x80),
						RGB(0x80, 0x80, 0), RGB(0xC0, 0xC0, 0xC0),
						RGB(0x80, 0x80, 0x80), RGB(0, 0, 0xFF), RGB(0, 0xFF, 0),
						RGB(0, 0xFF, 0xFF), RGB(0xFF, 0, 0), RGB(0xFF, 0, 0xFF),
						RGB(0xFF, 0xFF, 0), RGB(0xFF, 0xFF, 0xFF)
					};
					CHOOSECOLOR cc = { sizeof (CHOOSECOLOR) };
					cc.rgbResult = options[opBack];
					cc.Flags = CC_RGBINIT | CC_FULLOPEN;
					cc.lpCustColors = custColors;

					if (ChooseColor(&cc)) {
						options[opBack] = cc.rgbResult;
						DeleteObject((HGDIOBJ) SetClassLong(hWnd, GCL_HBRBACKGROUND,
								(LONG) CreateSolidBrush(options[opBack])));
						InvalidateRect(hWnd, NULL, TRUE);
					}
					return 0;
				}
				case IDM_OPTIONS_COLOR_BLOCK:
				{
					static COLORREF custColors[] = {
						RGB(0, 0, 0), RGB(0, 0, 0x80), RGB(0, 0x80, 0),
						RGB(0, 0x80, 0x80), RGB(0x80, 0, 0), RGB(0x80, 0, 0x80),
						RGB(0x80, 0x80, 0), RGB(0xC0, 0xC0, 0xC0),
						RGB(0x80, 0x80, 0x80), RGB(0, 0, 0xFF), RGB(0, 0xFF, 0),
						RGB(0, 0xFF, 0xFF), RGB(0xFF, 0, 0), RGB(0xFF, 0, 0xFF),
						RGB(0xFF, 0xFF, 0), RGB(0xFF, 0xFF, 0xFF)
					};
					CHOOSECOLOR cc = { sizeof (CHOOSECOLOR) };
					cc.rgbResult = options[opBlock];
					cc.Flags = CC_RGBINIT | CC_FULLOPEN;
					cc.lpCustColors = custColors;

					if (ChooseColor(&cc)) {
						options[opBlock] = cc.rgbResult;
						InvalidateRect(hWnd, NULL, TRUE);
					}
					return 0;
				}
				case IDM_OPTIONS_WINDOW_TOPMOST:
					options[opTopmost] ^= TRUE;
					CheckMenuItem(hMenu, IDM_OPTIONS_WINDOW_TOPMOST,
							options[opTopmost] ? MF_CHECKED : MF_UNCHECKED);
					SetWindowPos(hWnd, options[opTopmost] ? HWND_TOPMOST : HWND_NOTOPMOST,
							0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
					return 0;
				case IDM_OPTIONS_WINDOW_EASYMOVE:
					options[opEasyMove] ^= TRUE;
					CheckMenuItem(hMenu, IDM_OPTIONS_WINDOW_EASYMOVE,
							options[opEasyMove] ? MF_CHECKED : MF_UNCHECKED);
					return 0;
				case IDM_OPTIONS_KEYBINDINGS:
					if (DialogBoxParam((HINSTANCE) GetWindowLong(hWnd, GWL_HINSTANCE),
							"KeyBindings", hWnd, KeyBindingsProc, (LPARAM) accels)) {
						// ok was pressed, apply new accelerators
						DestroyAcceleratorTable(hAccel);
						hAccel = CreateAcceleratorTable(accels, NUM_ACCELS);
						refreshMenu(hMenu, accels, (BOOL) curFigure, fPause, options);
					}
					else
						// cancel was pressed, so cancel changes
						CopyAcceleratorTable(hAccel, accels, NUM_ACCELS);
					return 0;
				case IDM_OPTIONS_NEXT:
					options[opNext] ^= TRUE;
					CheckMenuItem(hMenu, IDM_OPTIONS_NEXT, options[opNext]
							? MF_CHECKED : MF_UNCHECKED);
					InvalidateRect(hWnd, NULL, TRUE);
					return 0;
				case IDM_OPTIONS_GRID:
					options[opGrid] ^= TRUE;
					CheckMenuItem(hMenu, IDM_OPTIONS_GRID, options[opGrid]
							? MF_CHECKED : MF_UNCHECKED);
					InvalidateRect(hWnd, NULL, TRUE);
					return 0;
				case IDM_OPTIONS_LAND:
					options[opLand] ^= TRUE;
					CheckMenuItem(hMenu, IDM_OPTIONS_LAND, options[opLand]
							? MF_CHECKED : MF_UNCHECKED);
					InvalidateRect(hWnd, NULL, TRUE);
					return 0;
				case IDM_OPTIONS_SAVE:
					fSave ^= TRUE;
					CheckMenuItem(hMenu, IDM_OPTIONS_SAVE, fSave
							? MF_CHECKED : MF_UNCHECKED);
					return 0;
				case IDM_HELP_README:
					if ((int) ShellExecute(NULL, "open", "readme.txt", NULL,
							NULL, SW_SHOWMAXIMIZED) == ERROR_FILE_NOT_FOUND)
						MessageBox(hWnd, " !!! !!!", "....",
								MB_ICONERROR | MB_OK);
					return 0;
				case IDM_HELP_ABOUT:
					MessageBox(hWnd,
							"Tetris clone by n/a, Christmas Edition ;)\n"
							"January - Febraury 2001\n"
							"http://www.chat.ru/~cynepnynep\n"
							"mailto:horny@mail.az\n\n"
							"If you like this, please send\n"
							"two bucks to Bill Gates\n"
							"Without windows we wouldn't\n"
							"see the sun (pun intended)",
							"About", MB_OK | MB_ICONQUESTION);
					return 0;
			}
			break;
		case WM_NCHITTEST:
			if (options[opEasyMove])
				if (DefWindowProc(hWnd, WM_NCHITTEST, wParam, lParam) == HTCLIENT)
					return HTCAPTION;
			break;
		case WM_RBUTTONDOWN:
		case WM_NCRBUTTONDOWN:
		{
			POINT point;
			point.x = LOWORD(lParam);
			point.y = HIWORD(lParam);
			if (uMsg == WM_RBUTTONDOWN)
				ClientToScreen(hWnd, &point);
			TrackPopupMenu(GetSubMenu(hMenu, 1), TPM_RIGHTBUTTON,
					point.x, point.y, 0, hWnd, NULL);
			return 0;
		}
		case WM_TIMER:
		{
			if (fPause)
				return 0;
			int x;
			int y;
			for (y = min(yPos + 2, NUM_ROWS - 1); y >= yPos; y--) {
				for (x = xPos; x < min(xPos + SIDE, NUM_COLS); x++)
					if ((cup[y][x] == 2) && (y == NUM_ROWS - 1 || cup[y + 1][x] == 1)) {
						// can't move down any more, so group current figure with others,
						// by making it static
						putBlockEx(cup, curFigure, xPos, yPos, btStatic);
						// find rows to crop, if any
						for (y = yPos; y <= min(yPos + 2, NUM_ROWS - 1); y++) {
							for (x = 0; x < NUM_COLS; x++)
								if (!cup[y][x])
									break;
							if (x < NUM_COLS)
								continue;
							for (int row = y; row; row--)
								for (x = 0; x < NUM_COLS; x++)
									cup[row][x] = cup[row - 1][x];
							for (x = 0; x < NUM_COLS; x++)
								cup[0][x] = 0;
							writeScore(hWnd, score += options[opSpeed], options[opSpeed]);
						}

						curFigure = nextFigure;
						nextFigure = figures[rand() % NUM_FIGURES];
						// check for game over
						for (y = 0; y < SIDE; y++)
						for (x = NUM_COLS / 2 - 1; x < (NUM_COLS / 2 - 1) + SIDE; x++)
							if (cup[y][x] && curFigure[y][x - xPos]) {
								fPause = TRUE;
								MessageBox(hWnd, "Game over", ":-(", MB_OK);
								SendMessage(hWnd, WM_COMMAND, IDM_GAME_NEW, 0);
								return 0;
							}
						// ok, next figure
						putBlockEx(cup, curFigure, xPos = NUM_COLS / 2 - 1,
								yPos = 0, btActive);
						InvalidateRect(hWnd, NULL, TRUE);
						return 0;
					}
			}
			// if we got here, we can move down
			putBlockEx(cup, curFigure, xPos, yPos, btZero);
			putBlockEx(cup, curFigure, xPos, ++yPos, btActive);

			RECT rect;
			SetRect(&rect, (xPos - 1) * 10, (yPos - 1) * 10,
					(xPos + SIDE) * 10, (yPos + SIDE) * 10);
			InvalidateRect(hWnd, &rect, TRUE);
			return 0;
		}
		case WM_PAINT:
		{
			PAINTSTRUCT ps;
			HDC hdc = BeginPaint(hWnd, &ps);
			// paint only if game started
			if (curFigure) {
				SelectObject(hdc, CreateSolidBrush(options[opBlock]));
				SelectObject(hdc, CreatePen(PS_SOLID, 1, options[opGrid]
						? options[opBack] : options[opBlock]));
				int x;
				int y;
				// update invalid rectangle
				for (y = ps.rcPaint.top / 10; y <= ps.rcPaint.bottom / 10; y++)
					for (x = ps.rcPaint.left / 10; x <= ps.rcPaint.right / 10; x++)
						if (cup[y][x])
							Rectangle(hdc, x * 10, y * 10, x * 10 + 10, y * 10 + 10);
				if (options[opNext])
					// draw next figure preview
					for (y = 0; y < SIDE; y++)
						for (x = 0; x < SIDE; x++)
							if (nextFigure[y][x])
								Rectangle(hdc, x * (10 / 2), y * (10 / 2),
										x * (10 / 2) + (10 / 2),
										y * (10 / 2) + (10 / 2));
				if (options[opLand]) {
					// draw landspot
					SetROP2(hdc, R2_NOT);
					for (x = xPos; x < xPos + SIDE; x++)
						for (y = yPos; y < yPos + SIDE; y++)
							if (curFigure[y - yPos][x - xPos]) {
								Rectangle(hdc, x * 10, NUM_ROWS * 10 - 3,
										x * 10 + 10, NUM_ROWS * 10);
								break;
							}
				}
				// cleanup...
				DeleteObject(SelectObject(hdc, GetStockObject(BLACK_PEN)));
				DeleteObject(SelectObject(hdc, GetStockObject(WHITE_BRUSH)));
			}
			EndPaint(hWnd, &ps);
			return 0;
		}
		case WM_CLOSE:
		{
			if (score > records[2].score) {
				// it's a record
				fPause = TRUE;
				DialogBoxParam((HINSTANCE) GetWindowLong(hWnd, GWL_HINSTANCE),
						"HighScores", hWnd, HighScoresProc, (LPARAM) records);
			}
			if (fSave) {
				HANDLE hFile = CreateFile("tetris.dat", GENERIC_WRITE, FILE_SHARE_WRITE,
						NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
				DWORD dwBytesWritten;
				WriteFile(hFile, options, sizeof options, &dwBytesWritten, NULL);
				WriteFile(hFile, records, sizeof records, &dwBytesWritten, NULL);
				WriteFile(hFile, accels, sizeof accels, &dwBytesWritten, NULL);
				CloseHandle(hFile);
			}
			else
				DeleteFile("tetris.dat");

			break;
		}
		case WM_DESTROY:
			// cleanup...
			KillTimer(hWnd, ID_TIMER);
			DeleteObject((HGDIOBJ) GetClassLong(hWnd, GCL_HBRBACKGROUND));
			DestroyAcceleratorTable(hAccel);
			PostQuitMessage(0);
			return 0;
	}
	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}

/************************************************************************************
*	HighScoresProc()
*		Display record scores. Can be called both by user and internally. In the first case
*		the records are meant for viewing only.
*	Parameters:
*		lParam == RECORDS* // array of records
*	Return value:
*		Ignored
*/
BOOL CALLBACK HighScoresProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	static int found = -1;
	static RECORDS* pRecords;

	switch (uMsg) {
		case WM_INITDIALOG:
		{
			HWND hEditWnd;
			pRecords = (RECORDS*) lParam;
			for (int i = 0; i < 3; i++) {
				hEditWnd = GetDlgItem(hDlg, IDC_NAME1 + i);
				SendMessage(hEditWnd, EM_SETLIMITTEXT, 30, 0);
				SetWindowText(hEditWnd, *pRecords[i].name ? pRecords[i].name : "<empty>");
				SetDlgItemInt(hDlg, IDC_SCORE1 + i, pRecords[i].score, FALSE);
				if (found == -1 && score > pRecords[i].score) {
					// insert new record here
					for (int ii = 2; ii > i; ii--) {
						// make room, discarding the least score
						lstrcpy(pRecords[ii].name, pRecords[ii - 1].name);
						pRecords[ii].score = pRecords[ii - 1].score;
					}
					SetDlgItemInt(hDlg, IDC_SCORE1 + i, pRecords[i].score = score, FALSE);
					SendMessage(hEditWnd, EM_SETREADONLY, FALSE, 0);
					SendMessage(hEditWnd, EM_SETSEL, 0, -1);
					SetFocus(hEditWnd);
					found = i;
				}
			}
			if (found == -1)
				// Not a record. Can only hold if the dialog was invoked by user.
				SetFocus(GetDlgItem(hDlg, IDOK));
			return FALSE;
		}
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDOK:
				case IDCANCEL:
					if (found != -1) {
						*(WORD*) pRecords[found].name = 30;
						pRecords[found].name[
								SendMessage(GetDlgItem(hDlg, IDC_NAME1 + found),
										EM_GETLINE, 0,
										(LPARAM) pRecords[found].name)
								] = '\0';	// ;)
						found = -1;
					}
					EndDialog(hDlg, TRUE);
					return TRUE;
			}
			break;
	}
	return FALSE;
}

/************************************************************************************
*	KeyBindingsProc()
*		Allows to view/reassign keys.
*	Parameters:
*		lParam == ACCEL* // array of current accelerator table
*	Return value:
*		TRUE if OK was pressed
*		FALSE if Cancel was pressed
*/
BOOL CALLBACK KeyBindingsProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	static char* actionNames[] = { // ListBox elements
		"GameNew", "GamePause", "GameHighScores", "GameExit",
		"ControlLeft", "ControlRight", "ControlDown", "ControlRotate",
		"ControlForceLeft", "ControlForceRight", "ControlForceDown", "OptionsSpeed1",
		"OptionsSpeed2", "OptionsSpeed3", "OptionsSpeed4", "OptionsColorBack",
		"OptionsColorBlock", "OptionsWindowAlwaysOnTop", "OptionsWindowEasyMove",
		"OptionsKeyBindings", "OptionsShowNextBlock", "OptionsShowGrid",
		"OptionsShowLandSpot", "HelpAbout"
	};
	static int itemIndex;	// current ListBox element (action names)
	static int keyIndex;	// current ComboBox element (key names)
	static ACCEL* accels;	// current accelerator table

	switch (uMsg) {
		case WM_INITDIALOG:
			accels = (ACCEL*) lParam;
			for (int i = 0; i < NUM_ACCELS; i++)
				// fill ListBox with action names
				SendDlgItemMessage(hDlg, IDL_KEYNAMES, LB_ADDSTRING, 0,
						(LPARAM) actionNames[i]);
			for (int i = 0; i < sizeof keys / sizeof *keys; i++)
				// fill ComboBox with key names
				SendDlgItemMessage(hDlg, IDC_KEYS, CB_ADDSTRING, 0, (LPARAM) keys[i]);

			SendDlgItemMessage(hDlg, IDL_KEYNAMES, LB_SETCURSEL, itemIndex, 0);
			SendMessage(hDlg, WM_COMMAND, MAKELONG(IDL_KEYNAMES, LBN_SELCHANGE),
					(LPARAM) GetDlgItem(hDlg, IDL_KEYNAMES));
			return TRUE;
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDL_KEYNAMES:
					switch (HIWORD(wParam)) {
						case LBN_SELCHANGE:
							itemIndex = SendDlgItemMessage(hDlg, IDL_KEYNAMES,
									LB_GETCURSEL, 0, 0);
							keyIndex = adjustKey(accels[itemIndex].key, TRUE);
							SendMessage(hDlg, WM_UPDATEDATA, 0, 0);
							return TRUE;
					}
					break;
				case IDC_KEYS:
					switch (HIWORD(wParam)) {
						case CBN_SELENDOK:
							keyIndex = SendDlgItemMessage(hDlg, IDC_KEYS, CB_GETCURSEL,
									0, 0);
							accels[itemIndex].key = adjustKey(keyIndex, FALSE);
							SendMessage(hDlg, WM_UPDATEDATA, 0, 0);
							return TRUE;
					}
					break;
				case IDC_CTRL:
				case IDC_SHIFT:
				case IDC_ALT:
					accels[itemIndex].fVirt ^= LOWORD(wParam) == IDC_CTRL ? FCONTROL
							: LOWORD(wParam) == IDC_SHIFT ? FSHIFT
							: FALT;
					SendMessage(hDlg, WM_UPDATEDATA, 0, 0);
					return TRUE;
				case IDDEFAULTS:
					if (MessageBox(hDlg, "Reset all key bindings to their"
							" factory defaults?", "Confirm", MB_YESNO
							| MB_ICONQUESTION | MB_DEFBUTTON2) == IDYES) {
						for (int i = 0; i < NUM_ACCELS; i++)
							accels[i] = defaultAccels[i];
						keyIndex = adjustKey(accels[itemIndex].key, TRUE);
						SetFocus(GetDlgItem(hDlg, IDOK));
						SendMessage(hDlg, WM_UPDATEDATA, 0, 0);
					}
					return TRUE;
				case IDOK:
				case IDCANCEL:
					EndDialog(hDlg, LOWORD(wParam) == IDOK);
					return TRUE;
			}
			break;
		case WM_UPDATEDATA:
			EnableWindow(GetDlgItem(hDlg, IDDEFAULTS), !isDefaultAccels(accels));
			EnableWindow(GetDlgItem(hDlg, IDOK), !isConflictingAccels(accels));
			SetWindowText(hDlg, IsWindowEnabled(GetDlgItem(hDlg, IDOK))
					? "Key Bindings" : "Key Bindings - there are conflicts");
			CheckDlgButton(hDlg, IDC_CTRL, accels[itemIndex].fVirt & FCONTROL
					? BST_CHECKED : BST_UNCHECKED);
			CheckDlgButton(hDlg, IDC_SHIFT, accels[itemIndex].fVirt & FSHIFT
					? BST_CHECKED : BST_UNCHECKED);
			CheckDlgButton(hDlg, IDC_ALT, accels[itemIndex].fVirt & FALT
					? BST_CHECKED : BST_UNCHECKED);
			SendDlgItemMessage(hDlg, IDC_KEYS, CB_SETCURSEL, keyIndex, 0);
			return TRUE;
	}
	return FALSE;
}

/************************************************************************************
*	putBlockEx
*		Place figure on board in specified coords.
*	Parameters:
*		cup[][NUM_COLS]		// game board
*		int (*block) [SIDE]	// figure to place
*		int xPos, yPos		// coordinates if figure, top-left corner
*		int BLOCK_TYPE bt	// can be 0 (btZero), 1 (btStatic), 2 (btActive)
*	Return value:
*		None
*/
void putBlockEx(int cup[][NUM_COLS], int (*block) [SIDE], int xPos, int yPos,
		BLOCK_TYPE bt) {
	for (int y = yPos; y < min(yPos + SIDE, NUM_ROWS); y++)
		for (int x = max(xPos, 0); x < min(xPos + SIDE, NUM_COLS); x++)
			if (block[y - yPos][x - xPos])
				cup[y][x] = (int) bt;
}

void writeScore(HWND hWnd, int score, int speed) {
	char s[50];
	wsprintf(s, "Score: %d   Speed: %d", score, speed);
	SetWindowText(hWnd, s);
}

/************************************************************************************
*	isDefaultAccels
*		Check if specified acceleartor table is same as default
*	Parameters:
*		accels[]	// accelerator table in question
*	Return value:
*		TRUE if specified acceleartor table is same as default
*		FALSE otherwise
*/
BOOL isDefaultAccels(const ACCEL accels[]) {
	for (int i = 0; i < NUM_ACCELS; i++)
		if ((accels[i].fVirt != defaultAccels[i].fVirt)
				|| (accels[i].key != defaultAccels[i].key))
			return FALSE;
	return TRUE;
}

/************************************************************************************
*	isConflictingAccels
*		Check if specified accelarators conflict with each other
*	Parameters:
*		accels[]	// accelerator table in question
*	Return value:
*		TRUE if specified accelarators conflict with each other
*		FALSE otherwise
*/
BOOL isConflictingAccels(const ACCEL accels[]) {
	for (int i = 0; i < NUM_ACCELS; i++)
		for (int j = 0; j < NUM_ACCELS; j++)
			if ((i != j) // do not compare to self
					&& (accels[i].fVirt == accels[j].fVirt)
					&& (accels[i].key == accels[j].key))
				return TRUE;
	return FALSE;
}

/************************************************************************************
*	adjustKey
*		Map between indices into array and their Virtual Keys
*	Parameters:
*		key					// either a Virtual Key, or an index into key array,
*							// depending on fMapToString
*		BOOL fMapToStrings	// if TRUE, key is a Virtual Key, else, it's an index into
*							// key array
*	Return value:
*		If fMapToString == TRUE, interpret key as a VK and return corresponding index into
*		keys array
*		Otherwise, interpret key as index into keys array and return corresponding VK
*/
int adjustKey(int key, BOOL fMapToStrings) {
	if (fMapToStrings)
		return key == VK_RETURN ? 0
				: key == VK_ESCAPE ? 1
				: key >= VK_SPACE && key <= VK_DOWN ? key - VK_SPACE + 2
				: key == VK_INSERT || key == VK_DELETE ? key - VK_INSERT + 11
				: key >= VK_0 && key <= VK_9 ? key - VK_0 + 13
				: key >= VK_A && key <= VK_Z ? key - VK_A + 23
				: key >= VK_NUMPAD0 && key <= VK_NUMPAD9 ? key - VK_NUMPAD0 + 49
				: key >= VK_F1 && key <= VK_F12 ? key - VK_F1 + 59
				: 0;
	return key == 0 ? VK_RETURN
			: key == 1 ? VK_ESCAPE
			: key >= 2 && key <= 10 ? key + VK_SPACE - 2
			: key == 11 || key == 12 ? key + VK_INSERT - 11
			: key >= 13 && key <= 22 ? key + VK_0 - 13
			: key >= 23 && key <= 48 ? key + VK_A - 23
			: key >= 49 && key <= 58 ? key + VK_NUMPAD0 - 49
			: key >= 59 && key <= 70 ? key + VK_F1 - 59
			: 0;
}

/************************************************************************************
*	refreshMenu
*		Redraw main menu.
*	Parameters:
*		accels[]	// current accelerator table
*		fCurFigure	// game started flag
*		fPause		// pause flag
*		options[]	// game options
*	Return value:
*		None
*/
void refreshMenu(HMENU hMenu, const ACCEL accels[], BOOL fCurFigure, BOOL fPause,
		const int options[]) {
	static char* menuItems[NUM_ACCELS] = {
		"&New Game\t", "&Pause\t", "&High Scores...\t", "E&xit\t",
		"&Left\t", "&Right\t", "&Down\t", "R&otate\t", "Force L&eft\t", "Force R&ight\t",
		"Force D&own\t", "Let Me &Relax\t", "&Normal\t", "Pretty &Fast\t",
		"This sucks! I &quit!\t", "&Background...\t", "B&lock...\t",
		"Always On &Top\t", "&Easy Move\t", "&Key Bindings...\t",
		"Show &Next\t", "Show &Grid\t", "Show &Land\t", "&About...\t"
	};
	char buf[50];

	for (int i = 0; i < NUM_ACCELS; i++) {
		lstrcpy(buf, (accels[i].cmd == IDM_GAME_PAUSE && fPause)
				? "Unpause\t" : menuItems[i]);
		lstrcat(buf, accels[i].fVirt & FCONTROL ? "Ctrl+" : "");
		lstrcat(buf, accels[i].fVirt & FSHIFT ? "Shift+" : "");
		lstrcat(buf, accels[i].fVirt & FALT ? "Alt+" : "");
		lstrcat(buf, keys[adjustKey(accels[i].key, TRUE)]);
		ModifyMenu(hMenu, accels[i].cmd, MF_STRING, accels[i].cmd, buf);
	}
	EnableMenuItem(hMenu, IDM_GAME_PAUSE, fCurFigure ? MF_ENABLED : MF_GRAYED);
	for (int i = IDM_CONTROL_LEFT; i <= IDM_CONTROL_FORCE_DOWN; i++)
		EnableMenuItem(hMenu, i, fCurFigure ? MF_ENABLED : MF_GRAYED);

	CheckMenuItem(hMenu, IDM_OPTIONS_WINDOW_TOPMOST, options[opTopmost]
			? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hMenu, IDM_OPTIONS_WINDOW_EASYMOVE, options[opEasyMove]
			? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hMenu, IDM_OPTIONS_NEXT, options[opNext] ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hMenu, IDM_OPTIONS_GRID, options[opGrid] ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hMenu, IDM_OPTIONS_LAND, options[opLand] ? MF_CHECKED : MF_UNCHECKED);
	CheckMenuRadioItem(hMenu, IDM_OPTIONS_SPEED_1, IDM_OPTIONS_SPEED_4,
			options[opSpeed] + IDM_OPTIONS_SPEED_1 - 1, MF_BYCOMMAND);
}
